#
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function

__metaclass__ = type

"""
The iosxr_ntp_global config file.
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to its desired end-state is
created.
"""

from copy import deepcopy

from ansible.module_utils.six import iteritems
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    dict_merge,
    dict_diff,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.resource_module import (
    ResourceModule,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import (
    Facts,
)
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.rm_templates.ntp_global import (
    Ntp_globalTemplate,
)


class Ntp_global(ResourceModule):
    """
    The iosxr_ntp_global config class
    """

    def __init__(self, module):
        super(Ntp_global, self).__init__(
            empty_fact_val={},
            facts_module=Facts(module),
            module=module,
            resource="ntp_global",
            tmplt=Ntp_globalTemplate(),
        )
        self.parsers = [
            "access_group.ipv4.peer",
            "access_group.ipv4.serve",
            "access_group.ipv4.serve_only",
            "access_group.ipv4.query_only",
            "access_group.ipv6.peer",
            "access_group.ipv6.serve",
            "access_group.ipv6.serve_only",
            "access_group.ipv6.query_only",
            "authenticate",
            "log_internal_sync",
            "broadcastdelay",
            "drift.aging_time",
            "drift.file",
            "ipv4.dscp",
            "ipv4.precedence",
            "ipv6.dscp",
            "ipv6.precedence",
            "max_associations",
            "master.stratum",
            "passive",
            "update_calendar",
            "source_interface",
        ]

    def execute_module(self):
        """Execute the module

        :rtype: A dictionary
        :returns: The result from module execution
        """
        if self.state not in ["parsed", "gathered"]:
            self.generate_commands()
            self.run_commands()
        return self.result

    def generate_commands(self):
        """Generate configuration commands to send based on
        want, have and desired state.
        """
        wantd = self._ntp_list_to_dict(self.want)
        haved = self._ntp_list_to_dict(self.have)

        # if state is merged, merge want onto have and then compare
        if self.state == "merged":
            wantd = dict_merge(haved, wantd)

        # if state is deleted, empty out wantd
        if self.state == "deleted":
            wantd = {}

        self._compare(want=wantd, have=haved)
        if self.state in ["overridden", "replaced"]:
            self.commands = [
                each for each in self.commands if "no" in each
            ] + [each for each in self.commands if "no" not in each]

    def _compare(self, want, have):
        """Leverages the base class `compare()` method and
        populates the list of commands to be run by comparing
        the `want` and `have` data with the `parsers` defined
        for the Ntp_global network resource.
        """
        self._compare_lists(want=want, have=have)
        self.compare(parsers=self.parsers, want=want, have=have)

    def _compare_lists(self, want, have):
        keys = [
            "authentication_keys",
            "peers",
            "servers",
            "trusted_keys",
            "interfaces",
            "source_vrfs",
            "access_group.vrfs",
        ]
        for x in keys:
            if "." in x:
                ag_vrf_list = x.split(".")
                wantx = want.get(ag_vrf_list[0], {}).get(ag_vrf_list[1], {})
                havex = have.get(ag_vrf_list[0], {}).get(ag_vrf_list[1], {})
                x = ag_vrf_list[1]
            else:
                wantx = want.get(x, {})
                havex = have.get(x, {})

            for wkey, wentry in iteritems(wantx):
                hentry = havex.pop(wkey, {})
                if wentry != hentry:
                    if x == "interfaces":
                        updates = dict_diff(hentry, wentry)
                        updates.update(name=wentry.get("name"))
                        updates.update(vrf=wentry.get("vrf", ""))
                        self.addcmd(updates, x)
                    else:
                        self.addcmd(wentry, x)

            # remove superfluos config
            for _hkey, hentry in iteritems(havex):
                if x == "interfaces":
                    if "vrf" in hentry:
                        self.commands.append(
                            "no ntp interface {0} vrf {1}".format(
                                _hkey.split("_")[0], _hkey.split("_")[1]
                            )
                        )
                    else:
                        self.commands.append(
                            "no ntp interface {0}".format(_hkey.split("_")[0])
                        )
                else:
                    self.addcmd(hentry, x, negate=True)

    def _ntp_list_to_dict(self, data):
        """Convert all list to dicts to dicts
        of dicts
        """
        tmp = deepcopy(data)
        if "access_group" in tmp:
            if "vrfs" in tmp["access_group"]:
                tmp["access_group"]["vrfs"] = {
                    i["name"]: i for i in tmp["access_group"]["vrfs"]
                }
        if "interfaces" in tmp:
            tmp["interfaces"] = {
                i["name"] + "_" + i.get("vrf", ""): i
                for i in tmp["interfaces"]
            }
        if "peers" in tmp:
            tmp["peers"] = {
                i["peer"] + "_" + i.get("vrf", ""): i for i in tmp["peers"]
            }
        if "servers" in tmp:
            tmp["servers"] = {
                i["server"] + "_" + i.get("vrf", ""): i for i in tmp["servers"]
            }

        pkey = {
            "authentication_keys": "id",
            "trusted_keys": "key_id",
            "source_vrfs": "vrf",
        }
        for k in pkey.keys():
            if k in tmp:
                tmp[k] = {i[pkey[k]]: i for i in tmp[k]}
        return tmp



#!/usr/bin/python
# This file is part of Ansible
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: route53_health_check
version_added: 1.0.0
short_description: Manage health-checks in Amazons Route53 DNS service
description:
  - Creates and deletes DNS Health checks in Amazons Route53 service.
  - Only the port, resource_path, string_match and request_interval are
    considered when updating existing health-checks.
options:
  state:
    description:
      - Specifies the action to take.
    choices: [ 'present', 'absent' ]
    type: str
    default: 'present'
  disabled:
    description:
      - Stops Route 53 from performing health checks.
      - See the AWS documentation for more details on the exact implications.
        U(https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/health-checks-creating-values.html)
      - Defaults to C(true) when creating a new health check.
    type: bool
    version_added: 2.1.0
  ip_address:
    description:
      - IP address of the end-point to check. Either this or I(fqdn) has to be provided.
      - IP addresses must be publicly routable.
    type: str
  port:
    description:
      - The port on the endpoint on which you want Amazon Route 53 to perform
        health checks. Required for TCP checks.
    type: int
  type:
    description:
      - The type of health check that you want to create, which indicates how
        Amazon Route 53 determines whether an endpoint is healthy.
    required: true
    choices: [ 'HTTP', 'HTTPS', 'HTTP_STR_MATCH', 'HTTPS_STR_MATCH', 'TCP' ]
    type: str
  resource_path:
    description:
      - The path that you want Amazon Route 53 to request when performing
        health checks. The path can be any value for which your endpoint will
        return an HTTP status code of 2xx or 3xx when the endpoint is healthy,
        for example the file /docs/route53-health-check.html.
      - Mutually exclusive with I(type='TCP').
      - The path must begin with a /
      - Maximum 255 characters.
    type: str
  fqdn:
    description:
      - Domain name of the endpoint to check. Either this or I(ip_address) has
        to be provided. When both are given the I(fqdn) is used in the C(Host:)
        header of the HTTP request.
    type: str
  string_match:
    description:
      - If the check type is HTTP_STR_MATCH or HTTP_STR_MATCH, the string
        that you want Amazon Route 53 to search for in the response body from
        the specified resource. If the string appears in the first 5120 bytes
        of the response body, Amazon Route 53 considers the resource healthy.
    type: str
  request_interval:
    description:
      - The number of seconds between the time that Amazon Route 53 gets a
        response from your endpoint and the time that it sends the next
        health-check request.
    default: 30
    choices: [ 10, 30 ]
    type: int
  failure_threshold:
    description:
      - The number of consecutive health checks that an endpoint must pass or
        fail for Amazon Route 53 to change the current status of the endpoint
        from unhealthy to healthy or vice versa.
      - Will default to C(3) if not specified on creation.
    choices: [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 ]
    type: int
  tags:
    description:
      - A hash/dictionary of tags to set on the health check.
    type: dict
    version_added: 2.1.0
  purge_tags:
    description:
      - Delete any tags not specified in I(tags).
    default: false
    type: bool
    version_added: 2.1.0
author: "zimbatm (@zimbatm)"
extends_documentation_fragment:
- amazon.aws.aws
- amazon.aws.ec2
'''

EXAMPLES = '''
- name: Create a health-check for host1.example.com and use it in record
  community.aws.route53_health_check:
    state: present
    fqdn: host1.example.com
    type: HTTP_STR_MATCH
    resource_path: /
    string_match: "Hello"
    request_interval: 10
    failure_threshold: 2
  register: my_health_check

- community.aws.route53:
    action: create
    zone: "example.com"
    type: CNAME
    record: "www.example.com"
    value: host1.example.com
    ttl: 30
    # Routing policy
    identifier: "host1@www"
    weight: 100
    health_check: "{{ my_health_check.health_check.id }}"

- name: Delete health-check
  community.aws.route53_health_check:
    state: absent
    fqdn: host1.example.com
'''

RETURN = r'''
health_check:
  description: Information about the health check.
  returned: success
  type: dict
  contains:
    action:
      description: The action performed by the module.
      type: str
      returned: When a change is or would be made.
      sample: 'updated'
    id:
      description: The Unique ID assigned by AWS to the health check.
      type: str
      returned: When the health check exists.
      sample: 50ec8a13-9623-4c66-9834-dd8c5aedc9ba
    health_check_version:
      description: The version number of the health check.
      type: int
      returned: When the health check exists.
      sample: 14
    health_check_config:
      description:
        - Detailed information about the health check.
        - May contain additional values from Route 53 health check
          features not yet supported by this module.
      type: dict
      returned: When the health check exists.
      contains:
        type:
          description: The type of the health check.
          type: str
          returned: When the health check exists.
          sample: 'HTTPS_STR_MATCH'
        failure_threshold:
          description:
            - The number of consecutive health checks that an endpoint must pass or fail for Amazon Route 53 to
              change the current status of the endpoint from unhealthy to healthy or vice versa.
          type: int
          returned: When the health check exists.
          sample: 3
        fully_qualified_domain_name:
          description: The FQDN configured for the health check to test.
          type: str
          returned: When the health check exists and an FQDN is configured.
          sample: 'updated'
        ip_address:
          description: The IPv4 or IPv6 IP address of the endpoint to be queried.
          type: str
          returned: When the health check exists and a specific IP address is configured.
          sample: ''
        port:
          description: The port on the endpoint that the health check will query.
          type: str
          returned: When the health check exists.
          sample: 'updated'
        request_interval:
          description: The number of seconds between health check queries.
          type: int
          returned: When the health check exists.
          sample: 30
        resource_path:
          description: The URI path to query when performing an HTTP/HTTPS based health check.
          type: str
          returned: When the health check exists and a resource path has been configured.
          sample: '/healthz'
        search_string:
          description: A string that must be present in the response for a health check to be considered successful.
          type: str
          returned: When the health check exists and a search string has been configured.
          sample: 'ALIVE'
        disabled:
          description: Whether the health check has been disabled or not.
          type: bool
          returned: When the health check exists.
          sample: false
    tags:
      description: A dictionary representing the tags on the health check.
      type: dict
      returned: When the health check exists.
      sample: '{"my_key": "my_value"}'
'''

import uuid

try:
    import botocore
except ImportError:
    pass  # Handled by HAS_BOTO

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.core import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import AWSRetry
from ansible_collections.community.aws.plugins.module_utils.route53 import get_tags
from ansible_collections.community.aws.plugins.module_utils.route53 import manage_tags


def _list_health_checks(**params):
    try:
        results = client.list_health_checks(aws_retry=True, **params)
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, msg='Failed to list health checks')
    return results


def find_health_check(ip_addr, fqdn, hc_type, request_interval, port):
    """Searches for health checks that have the exact same set of immutable values"""

    # In lieu of an Id we perform matches against the following values:
    # - ip_addr
    # - fqdn
    # - type (immutable)
    # - request_interval
    # - port

    # Because the list and route53 provides no 'filter' mechanism,
    # the using a paginator would result in (on average) double the
    # number of API calls and can get really slow.
    # Additionally, we can't properly wrap the paginator, so retrying means
    # starting from scratch with a paginator
    results = _list_health_checks()

    while True:
        for check in results.get('HealthChecks'):
            config = check.get('HealthCheckConfig')
            if (
                config.get('IPAddress', None) == ip_addr and
                config.get('FullyQualifiedDomainName', None) == fqdn and
                config.get('Type') == hc_type and
                config.get('RequestInterval') == request_interval and
                config.get('Port', None) == port
            ):
                return check

        if results.get('IsTruncated', False):
            results = _list_health_checks(Marker=results.get('NextMarker'))
        else:
            return None


def delete_health_check(check_id):
    if not check_id:
        return False, None

    if module.check_mode:
        return True, 'delete'

    try:
        client.delete_health_check(
            aws_retry=True,
            HealthCheckId=check_id,
        )
    except is_boto3_error_code('NoSuchHealthCheck'):
        # Handle the deletion race condition as cleanly as possible
        return False, None
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg='Failed to list health checks')

    return True, 'delete'


def create_health_check(ip_addr_in, fqdn_in, type_in, request_interval_in, port_in):

    # In general, if a request is repeated with the same CallerRef it won't
    # result in a duplicate check appearing.  This means we can safely use our
    # retry decorators
    caller_ref = str(uuid.uuid4())
    missing_args = []

    health_check = dict(
        Type=type_in,
        RequestInterval=request_interval_in,
        Port=port_in,
    )
    if module.params.get('disabled') is not None:
        health_check['Disabled'] = module.params.get('disabled')
    if ip_addr_in:
        health_check['IPAddress'] = ip_addr_in
    if fqdn_in:
        health_check['FullyQualifiedDomainName'] = fqdn_in

    if type_in in ['HTTP', 'HTTPS', 'HTTP_STR_MATCH', 'HTTPS_STR_MATCH']:
        resource_path = module.params.get('resource_path')
        # if not resource_path:
        #     missing_args.append('resource_path')
        if resource_path:
            health_check['ResourcePath'] = resource_path
    if type_in in ['HTTP_STR_MATCH', 'HTTPS_STR_MATCH']:
        string_match = module.params.get('string_match')
        if not string_match:
            missing_args.append('string_match')
        health_check['SearchString'] = module.params.get('string_match')

    failure_threshold = module.params.get('failure_threshold')
    if not failure_threshold:
        failure_threshold = 3
    health_check['FailureThreshold'] = failure_threshold

    if missing_args:
        module.fail_json(msg='missing required arguments for creation: {0}'.format(
            ', '.join(missing_args)),
        )

    if module.check_mode:
        return True, 'create', None

    try:
        result = client.create_health_check(
            aws_retry=True,
            CallerReference=caller_ref,
            HealthCheckConfig=health_check,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg='Failed to create health check.', health_check=health_check)

    check_id = result.get('HealthCheck').get('Id')
    return True, 'create', check_id


def update_health_check(existing_check):
    # In theory it's also possible to update the IPAddress, Port and
    # FullyQualifiedDomainName, however, because we use these in lieu of a
    # 'Name' to uniquely identify the health check this isn't currently
    # supported.  If we accepted an ID it would be possible to modify them.

    changes = dict()
    existing_config = existing_check.get('HealthCheckConfig')

    resource_path = module.params.get('resource_path', None)
    if resource_path and resource_path != existing_config.get('ResourcePath'):
        changes['ResourcePath'] = resource_path

    search_string = module.params.get('string_match', None)
    if search_string and search_string != existing_config.get('SearchString'):
        changes['SearchString'] = search_string

    failure_threshold = module.params.get('failure_threshold', None)
    if failure_threshold and failure_threshold != existing_config.get('FailureThreshold'):
        changes['FailureThreshold'] = failure_threshold

    disabled = module.params.get('disabled', None)
    if disabled is not None and disabled != existing_config.get('Disabled'):
        changes['Disabled'] = module.params.get('disabled')

    # No changes...
    if not changes:
        return False, None

    if module.check_mode:
        return True, 'update'

    check_id = existing_check.get('Id')
    # This makes sure we're starting from the version we think we are...
    version_id = existing_check.get('HealthCheckVersion', 1)
    try:
        client.update_health_check(
            HealthCheckId=check_id,
            HealthCheckVersion=version_id,
            **changes,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg='Failed to update health check.', id=check_id)

    return True, 'update'


def describe_health_check(id):
    if not id:
        return dict()

    try:
        result = client.get_health_check(
            aws_retry=True,
            HealthCheckId=id,
        )
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(e, msg='Failed to get health check.', id=id)

    health_check = result.get('HealthCheck', {})
    health_check = camel_dict_to_snake_dict(health_check)
    tags = get_tags(module, client, 'healthcheck', id)
    health_check['tags'] = tags
    return health_check


def main():
    argument_spec = dict(
        state=dict(choices=['present', 'absent'], default='present'),
        disabled=dict(type='bool'),
        ip_address=dict(),
        port=dict(type='int'),
        type=dict(required=True, choices=['HTTP', 'HTTPS', 'HTTP_STR_MATCH', 'HTTPS_STR_MATCH', 'TCP']),
        resource_path=dict(),
        fqdn=dict(),
        string_match=dict(),
        request_interval=dict(type='int', choices=[10, 30], default=30),
        failure_threshold=dict(type='int', choices=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
        tags=dict(type='dict'),
        purge_tags=dict(type='bool', default=False),
    )

    args_one_of = [
        ['ip_address', 'fqdn'],
    ]

    args_if = [
        ['type', 'TCP', ('port',)],
    ]

    global module
    global client

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        required_one_of=args_one_of,
        required_if=args_if,
        supports_check_mode=True,
    )

    state_in = module.params.get('state')
    ip_addr_in = module.params.get('ip_address')
    port_in = module.params.get('port')
    type_in = module.params.get('type')
    resource_path_in = module.params.get('resource_path')
    fqdn_in = module.params.get('fqdn')
    string_match_in = module.params.get('string_match')
    request_interval_in = module.params.get('request_interval')
    failure_threshold_in = module.params.get('failure_threshold')

    # Default port
    if port_in is None:
        if type_in in ['HTTP', 'HTTP_STR_MATCH']:
            port_in = 80
        elif type_in in ['HTTPS', 'HTTPS_STR_MATCH']:
            port_in = 443

    if string_match_in:
        if type_in not in ['HTTP_STR_MATCH', 'HTTPS_STR_MATCH']:
            module.fail_json(msg="parameter 'string_match' argument is only for the HTTP(S)_STR_MATCH types")
        if len(string_match_in) > 255:
            module.fail_json(msg="parameter 'string_match' is limited to 255 characters max")

    client = module.client('route53', retry_decorator=AWSRetry.jittered_backoff())

    changed = False
    action = None
    check_id = None

    existing_check = find_health_check(ip_addr_in, fqdn_in, type_in, request_interval_in, port_in)

    if existing_check:
        check_id = existing_check.get('Id')

    if state_in == 'absent':
        changed, action = delete_health_check(check_id)
        check_id = None
    elif state_in == 'present':
        if existing_check is None:
            changed, action, check_id = create_health_check(ip_addr_in, fqdn_in, type_in, request_interval_in, port_in)
        else:
            changed, action = update_health_check(existing_check)
        if check_id:
            changed |= manage_tags(module, client, 'healthcheck', check_id,
                                   module.params.get('tags'), module.params.get('purge_tags'))

    health_check = describe_health_check(id=check_id)
    health_check['action'] = action
    module.exit_json(
        changed=changed,
        health_check=health_check,
    )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

from ansible_collections.community.internal_test_tools.tests.unit.utils.fetch_url_module_framework import (
    BaseTestModule,
    FetchUrlCall,
)

from ansible_collections.community.dns.plugins.modules import hetzner_dns_record_set

# These imports are needed so patching below works
import ansible_collections.community.dns.plugins.module_utils.http  # noqa

from .hetzner import (
    HETZNER_JSON_DEFAULT_ENTRIES,
    HETZNER_JSON_ZONE_GET_RESULT,
    HETZNER_JSON_ZONE_LIST_RESULT,
    HETZNER_JSON_ZONE_RECORDS_GET_RESULT,
)


class TestHetznerDNSRecordJSON(BaseTestModule):
    MOCK_ANSIBLE_MODULEUTILS_BASIC_ANSIBLEMODULE = 'ansible_collections.community.dns.plugins.modules.hetzner_dns_record_set.AnsibleModule'
    MOCK_ANSIBLE_MODULEUTILS_URLS_FETCH_URL = 'ansible_collections.community.dns.plugins.module_utils.http.fetch_url'

    def test_unknown_zone(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.org',
            'record': 'example.org',
            'type': 'MX',
            'ttl': 3600,
            'value': [
                '10 example.com',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.org')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
        ])

        assert result['msg'] == 'Zone not found'

    def test_unknown_zone_id(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_id': '23',
            'record': 'example.org',
            'type': 'MX',
            'ttl': 3600,
            'value': [
                '10 example.com',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 404)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones/23')
            .return_header('Content-Type', 'application/json')
            .result_json({'error': {'message': 'zone not found', 'code': 404}}),
        ])

        assert result['msg'] == 'Zone not found'

    def test_unknown_zone_id_prefix(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_id': '23',
            'prefix': '',
            'type': 'MX',
            'ttl': 3600,
            'value': [
                '10 example.com',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 404)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '23')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json({'records': [], 'error': {'message': 'zone not found', 'code': 404}}),
        ])

        assert result['msg'] == 'Zone not found'

    def test_auth_error(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.org',
            'record': 'example.org',
            'type': 'MX',
            'ttl': 3600,
            'value': [
                '10 example.com',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 401)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.org')
            .result_json({'message': 'Invalid authentication credentials'}),
        ])

        assert result['msg'] == (
            'Cannot authenticate: Unauthorized: the authentication parameters are incorrect (HTTP status 401): Invalid authentication credentials'
        )

    def test_other_error(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.org',
            'record': 'example.org',
            'type': 'MX',
            'ttl': 3600,
            'value': [
                '10 example.com',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 500)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.org')
            .result_str(''),
        ])

        assert result['msg'].startswith('Error: GET https://dns.hetzner.com/api/v1/zones?')
        assert 'did not yield JSON data, but HTTP status code 500 with Content-Type' in result['msg']

    def test_conversion_error(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'TXT',
            'ttl': 3600,
            'value': [
                u'"hellö',
            ],
            'txt_transformation': 'quoted',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['msg'] == (
            'Error while converting DNS values: While processing record from the user: Missing double quotation mark at the end of value'
        )

    def test_idempotency_present(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'MX',
            'ttl': 3600,
            'value': [
                '10 example.com',
            ],
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'
        assert result['diff']['before'] == {
            'record': 'example.com',
            'prefix': '',
            'type': 'MX',
            'ttl': 3600,
            'value': ['10 example.com'],
        }
        assert result['diff']['before'] == result['diff']['after']

    def test_idempotency_absent_value(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'record': '*.example.com',
            'type': 'A',
            'ttl': 3600,
            'value': [
                '1.2.3.6',
            ],
            'on_existing': 'keep',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'
        assert result['diff']['before'] == {
            'record': '*.example.com',
            'prefix': '*',
            'type': 'A',
            'ttl': 3600,
            'value': ['1.2.3.5'],
        }
        assert result['diff']['before'] == result['diff']['after']

    def test_idempotency_absent_value_prefix(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'prefix': '*',
            'type': 'A',
            'ttl': 3600,
            'value': [
                '1.2.3.6',
            ],
            'on_existing': 'keep',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'

    def test_idempotency_absent_ttl(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'record': '*.example.com',
            'type': 'A',
            'ttl': 1800,
            'value': [
                '1.2.3.5',
            ],
            'on_existing': 'keep',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'

    def test_idempotency_absent_type(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'CAA',
            'ttl': 3600,
            'value': [
                '0 issue "letsencrypt.org"',
            ],
            'on_existing': 'keep',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'

    def test_idempotency_absent_record(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com.',
            'record': 'somewhere.example.com.',
            'type': 'A',
            'ttl': 3600,
            'value': [
                '1.2.3.6',
            ],
            'on_existing': 'keep',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'
        assert 'warnings' not in result

    def test_idempotency_absent_record_warn(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com.',
            'record': 'somewhere.example.com.',
            'type': 'A',
            'ttl': 3600,
            'value': [
                '1.2.3.6',
            ],
            'on_existing': 'keep_and_warn',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'
        assert list(result['warnings']) == ["Record already exists with different value. Set on_existing=replace to remove it"]

    def test_idempotency_absent_record_fail(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com.',
            'record': 'somewhere.example.com.',
            'type': 'A',
            'ttl': 3600,
            'value': [
                '1.2.3.6',
            ],
            'on_existing': 'keep_and_fail',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['msg'] == "Record already exists with different value. Set on_existing=replace to remove it"

    def test_absent(self, mocker):
        record = HETZNER_JSON_DEFAULT_ENTRIES[0]
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'record': ((record['name'] + '.') if record['name'] != '@' else '') + 'example.com',
            'type': record['type'],
            'ttl': record['ttl'],
            'value': [
                record['value'],
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('DELETE', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/{0}'.format(record['id']))
            .result_str(''),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_absent_error(self, mocker):
        record = HETZNER_JSON_DEFAULT_ENTRIES[0]
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'record': ((record['name'] + '.') if record['name'] != '@' else '') + 'example.com',
            'type': record['type'],
            'ttl': record['ttl'],
            'value': [
                record['value'],
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('DELETE', 500)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/{0}'.format(record['id']))
            .return_header('Content-Type', 'application/json')
            .result_json({'error': {'message': 'Internal Server Error', 'code': 500}}),
        ])

        print(result['msg'])
        assert result['msg'] == (
            'Error: Expected HTTP status 200, 404 for DELETE https://dns.hetzner.com/api/v1/records/125,'
            ' but got HTTP status 500 (Internal Server Error) with error message "Internal Server Error" (error code 500)'
        )

    def test_absent_bulk(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'value': [],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('DELETE', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/130')
            .result_str(''),
            FetchUrlCall('DELETE', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/131')
            .result_str(''),
            # Record 132 has been deleted between querying and we trying to delete it
            FetchUrlCall('DELETE', 404)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/132')
            .return_header('Content-Type', 'application/json')
            .result_json({'message': 'record does not exist'}),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_absent_bulk_error(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'value': [],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('DELETE', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/130')
            .result_str(''),
            FetchUrlCall('DELETE', 500)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/131')
            .return_header('Content-Type', 'application/json')
            .result_json({'error': {'message': 'Internal Server Error', 'code': 500}}),
        ])

        assert result['msg'] == (
            'Error: Expected HTTP status 200, 404 for DELETE https://dns.hetzner.com/api/v1/records/131,'
            ' but got HTTP status 500 (Internal Server Error) with error message "Internal Server Error" (error code 500)'
        )

    def test_absent_other_value(self, mocker):
        record = HETZNER_JSON_DEFAULT_ENTRIES[0]
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'absent',
            'zone_name': 'example.com',
            'record': ((record['name'] + '.') if record['name'] != '@' else '') + 'example.com',
            'type': record['type'],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('DELETE', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/{0}'.format(record['id']))
            .result_str(''),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_change_add_one_check_mode(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_id': '42',
            'record': 'example.com',
            'type': 'CAA',
            'ttl': 3600,
            'value': [
                '0 issue "letsencrypt.org"',
            ],
            '_ansible_check_mode': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones/42')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_GET_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_change_add_one_check_mode_prefix(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_id': '42',
            'prefix': '@',
            'type': 'CAA',
            'ttl': 3600,
            'value': [
                '0 issue "letsencrypt.org"',
            ],
            '_ansible_diff': True,
            '_ansible_check_mode': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {}
        assert result['diff']['after'] == {
            'prefix': '',
            'type': 'CAA',
            'ttl': 3600,
            'value': ['0 issue "letsencrypt.org"'],
        }

    def test_change_add_one(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'CAA',
            'ttl': 3600,
            'value': [
                '128 issue "letsencrypt.org xxx"',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('POST', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'CAA')
            .expect_json_value(['ttl'], 3600)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], '128 issue "letsencrypt.org xxx"')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '133',
                    'type': 'CAA',
                    'name': '@',
                    'value': '128 issue "letsencrypt.org xxx"',
                    'ttl': 3600,
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_change_add_one_prefix(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'prefix': '',
            'type': 'CAA',
            'ttl': 3600,
            'value': [
                '128 issue "letsencrypt.org"',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('POST', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'CAA')
            .expect_json_value(['ttl'], 3600)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], '128 issue "letsencrypt.org"')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '133',
                    'type': 'CAA',
                    'name': '@',
                    'value': '128 issue "letsencrypt.org"',
                    'ttl': 3600,
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_change_add_one_idn_prefix(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'prefix': '☺',
            'type': 'CAA',
            'ttl': 3600,
            'value': [
                '128 issue "letsencrypt.org"',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('POST', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'CAA')
            .expect_json_value(['ttl'], 3600)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], 'xn--74h')
            .expect_json_value(['value'], '128 issue "letsencrypt.org"')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '133',
                    'type': 'CAA',
                    'name': 'xn--74h',
                    'value': '128 issue "letsencrypt.org"',
                    'ttl': 3600,
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'

    def test_change_modify_list_fail(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': None,
            'value': [
                'helium.ns.hetzner.de.',
                'ytterbium.ns.hetzner.com.',
            ],
            'on_existing': 'keep_and_fail',
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['msg'] == "Record already exists with different value. Set on_existing=replace to replace it"

    def test_change_modify_list_warn(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': 10800,
            'value': [
                'helium.ns.hetzner.de.',
                'ytterbium.ns.hetzner.com.',
            ],
            'on_existing': 'keep_and_warn',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert result['changed'] is False
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {
            'record': 'example.com',
            'prefix': '',
            'type': 'NS',
            'ttl': None,
            'value': ['helium.ns.hetzner.de.', 'hydrogen.ns.hetzner.com.', 'oxygen.ns.hetzner.com.'],
        }
        assert result['diff']['after'] == result['diff']['before']
        assert list(result['warnings']) == ["Record already exists with different value. Set on_existing=replace to replace it"]

    def test_change_modify_list_keep(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': None,
            'value': [
                'helium.ns.hetzner.de.',
                'ytterbium.ns.hetzner.com.',
            ],
            'on_existing': 'keep',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
        ])

        assert 'warnings' not in result
        assert result['changed'] is False
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {
            'record': 'example.com',
            'prefix': '',
            'type': 'NS',
            'ttl': None,
            'value': ['helium.ns.hetzner.de.', 'hydrogen.ns.hetzner.com.', 'oxygen.ns.hetzner.com.'],
        }
        assert result['diff']['after'] == result['diff']['before']

    def test_change_modify_list(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': None,
            'value': [
                'helium.ns.hetzner.de.',
                'ytterbium.ns.hetzner.com.',
            ],
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('DELETE', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/131')
            .result_str(''),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/132')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value_absent(['ttl'])
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'ytterbium.ns.hetzner.com.')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '132',
                    'type': 'NS',
                    'name': '@',
                    'value': 'ytterbium.ns.hetzner.com.',
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {
            'record': 'example.com',
            'prefix': '',
            'type': 'NS',
            'ttl': None,
            'value': ['helium.ns.hetzner.de.', 'hydrogen.ns.hetzner.com.', 'oxygen.ns.hetzner.com.'],
        }
        assert result['diff']['after'] == {
            'record': 'example.com',
            'prefix': '',
            'type': 'NS',
            'ttl': None,
            'value': ['helium.ns.hetzner.de.', 'ytterbium.ns.hetzner.com.'],
        }

    def test_change_modify_txt_unquoted(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'foo.example.com',
            'type': 'TXT',
            'ttl': None,
            'value': [u'bär "with quotes" (use \\ to escape)!'],
            'txt_transformation': 'unquoted',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/201')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'TXT')
            .expect_json_value_absent(['ttl'])
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], 'foo')
            .expect_json_value(['value'], u'"bär \\"with quotes\\" (use \\\\ to escape)!"')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '201',
                    'type': 'TXT',
                    'name': 'foo',
                    'value': u'"bär \\"with quotes\\" (use \\\\ to escape)!"',
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {
            'record': 'foo.example.com',
            'prefix': 'foo',
            'type': 'TXT',
            'ttl': None,
            'value': [u'bär "with quotes" (use \\ to escape)'],
        }
        assert result['diff']['after'] == {
            'record': 'foo.example.com',
            'prefix': 'foo',
            'type': 'TXT',
            'ttl': None,
            'value': [u'bär "with quotes" (use \\ to escape)!'],
        }

    def test_change_modify_txt_quoted(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'foo.example.com',
            'type': 'TXT',
            'ttl': None,
            'value': [r'"b\303\244r \"with quotes\" (use \\ to escape)!"'],
            'txt_transformation': 'quoted',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/201')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'TXT')
            .expect_json_value_absent(['ttl'])
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], 'foo')
            .expect_json_value(['value'], u'"bär \\"with quotes\\" (use \\\\ to escape)!"')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '201',
                    'type': 'TXT',
                    'name': 'foo',
                    'value': u'"bär \\"with quotes\\" (use \\\\ to escape)!"',
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {
            'record': 'foo.example.com',
            'prefix': 'foo',
            'type': 'TXT',
            'ttl': None,
            'value': [r'"b\303\244r \"with quotes\" (use \\ to escape)"'],
        }
        assert result['diff']['after'] == {
            'record': 'foo.example.com',
            'prefix': 'foo',
            'type': 'TXT',
            'ttl': None,
            'value': [r'"b\303\244r \"with quotes\" (use \\ to escape)!"'],
        }

    def test_change_modify_txt_api(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'foo.example.com',
            'type': 'TXT',
            'ttl': None,
            'value': [u'bär " " \\"with " " quotes\\" " (use \\\\ to escape)!"'],
            'txt_transformation': 'api',
            '_ansible_diff': True,
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/201')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'TXT')
            .expect_json_value_absent(['ttl'])
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], 'foo')
            .expect_json_value(['value'], u'bär " " \\"with " " quotes\\" " (use \\\\ to escape)!"')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '201',
                    'type': 'TXT',
                    'name': 'foo',
                    'value': u'bär " " \\"with " " quotes\\" " (use \\\\ to escape)!"',
                    'zone_id': '42',
                },
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'
        assert 'diff' in result
        assert 'before' in result['diff']
        assert 'after' in result['diff']
        assert result['diff']['before'] == {
            'record': 'foo.example.com',
            'prefix': 'foo',
            'type': 'TXT',
            'ttl': None,
            'value': [u'bär " \\"with quotes\\"" " " "(use \\\\ to escape)"'],
        }
        assert result['diff']['after'] == {
            'record': 'foo.example.com',
            'prefix': 'foo',
            'type': 'TXT',
            'ttl': None,
            'value': [u'bär " " \\"with " " quotes\\" " (use \\\\ to escape)!"'],
        }

    def test_change_modify_bulk(self, mocker):
        result = self.run_module_success(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': 10800,
            'value': [
                'a1',
                'a2',
                'a3',
                'a4',
                'a5',
                'a6',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/132')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a1')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '132',
                    'type': 'NS',
                    'name': '@',
                    'value': 'a1',
                    'ttl': 10800,
                    'zone_id': '42',
                },
            }),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/131')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a2')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '131',
                    'type': 'NS',
                    'name': '@',
                    'value': 'a2',
                    'ttl': 10800,
                    'zone_id': '42',
                },
            }),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/130')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a3')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '130',
                    'type': 'NS',
                    'name': '@',
                    'value': 'a3',
                    'ttl': 10800,
                    'zone_id': '42',
                },
            }),
            FetchUrlCall('POST', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/bulk')
            .expect_json_value_absent(['records', 0, 'id'])
            .expect_json_value(['records', 0, 'type'], 'NS')
            .expect_json_value(['records', 0, 'ttl'], 10800)
            .expect_json_value(['records', 0, 'zone_id'], '42')
            .expect_json_value(['records', 0, 'name'], '@')
            .expect_json_value(['records', 0, 'value'], 'a4')
            .expect_json_value_absent(['records', 1, 'id'])
            .expect_json_value(['records', 1, 'type'], 'NS')
            .expect_json_value(['records', 1, 'ttl'], 10800)
            .expect_json_value(['records', 1, 'zone_id'], '42')
            .expect_json_value(['records', 1, 'name'], '@')
            .expect_json_value(['records', 1, 'value'], 'a5')
            .expect_json_value_absent(['records', 2, 'id'])
            .expect_json_value(['records', 2, 'type'], 'NS')
            .expect_json_value(['records', 2, 'ttl'], 10800)
            .expect_json_value(['records', 2, 'zone_id'], '42')
            .expect_json_value(['records', 2, 'name'], '@')
            .expect_json_value(['records', 2, 'value'], 'a6')
            .expect_json_value_absent(['records', 3])
            .return_header('Content-Type', 'application/json')
            .result_json({
                'invalid_records': [],
                'valid_records': [],
                'records': [
                    {
                        'id': '300',
                        'type': 'NS',
                        'name': '@',
                        'value': 'a4',
                        'ttl': 10800,
                        'zone_id': '42',
                    },
                    {
                        'id': '301',
                        'type': 'NS',
                        'name': '@',
                        'value': 'a5',
                        'ttl': 10800,
                        'zone_id': '42',
                    },
                    {
                        'id': '302',
                        'type': 'NS',
                        'name': '@',
                        'value': 'a6',
                        'ttl': 10800,
                        'zone_id': '42',
                    },
                ],
            }),
        ])

        assert result['changed'] is True
        assert result['zone_id'] == '42'
        assert 'diff' not in result

    def test_change_modify_bulk_errors(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': 10800,
            'value': [
                'a1',
                'a2',
                'a3',
                'a4',
                'a5',
                'a6',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 500)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/132')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a1')
            .return_header('Content-Type', 'application/json')
            .result_json({'message': 'Internal Server Error'}),
        ])

        assert result['msg'] == (
            'Error: Expected HTTP status 200, 422 for PUT https://dns.hetzner.com/api/v1/records/132,'
            ' but got HTTP status 500 (Internal Server Error) with message "Internal Server Error"'
        )

    def test_change_modify_bulk_errors_2(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'NS',
            'ttl': 10800,
            'value': [
                'a1',
                'a2',
                'a3',
                'a4',
                'a5',
                'a6',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/132')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a1')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '132',
                    'type': 'NS',
                    'name': '@',
                    'value': 'a1',
                    'ttl': 10800,
                    'zone_id': '42',
                },
            }),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/131')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a2')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '131',
                    'type': 'NS',
                    'name': '@',
                    'value': 'a2',
                    'ttl': 10800,
                    'zone_id': '42',
                },
            }),
            FetchUrlCall('PUT', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/130')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'NS')
            .expect_json_value(['ttl'], 10800)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], 'a3')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '130',
                    'type': 'NS',
                    'name': '@',
                    'value': 'a3',
                    'ttl': 10800,
                    'zone_id': '42',
                },
            }),
            FetchUrlCall('POST', 422)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/bulk')
            .expect_json_value_absent(['records', 0, 'id'])
            .expect_json_value(['records', 0, 'type'], 'NS')
            .expect_json_value(['records', 0, 'ttl'], 10800)
            .expect_json_value(['records', 0, 'zone_id'], '42')
            .expect_json_value(['records', 0, 'name'], '@')
            .expect_json_value(['records', 0, 'value'], 'a4')
            .expect_json_value_absent(['records', 1, 'id'])
            .expect_json_value(['records', 1, 'type'], 'NS')
            .expect_json_value(['records', 1, 'ttl'], 10800)
            .expect_json_value(['records', 1, 'zone_id'], '42')
            .expect_json_value(['records', 1, 'name'], '@')
            .expect_json_value(['records', 1, 'value'], 'a5')
            .expect_json_value_absent(['records', 2, 'id'])
            .expect_json_value(['records', 2, 'type'], 'NS')
            .expect_json_value(['records', 2, 'ttl'], 10800)
            .expect_json_value(['records', 2, 'zone_id'], '42')
            .expect_json_value(['records', 2, 'name'], '@')
            .expect_json_value(['records', 2, 'value'], 'a6')
            .expect_json_value_absent(['records', 3])
            .return_header('Content-Type', 'application/json')
            .result_json({
                'invalid_records': [
                    {
                        'type': 'NS',
                        'name': '@',
                        'value': 'a4',
                        'ttl': 10800,
                        'zone_id': '42',
                    },
                    {
                        'type': 'NS',
                        'name': '@',
                        'value': 'a5',
                        'ttl': 10800,
                        'zone_id': '42',
                    },
                ],
                'valid_records': [
                    {
                        'type': 'NS',
                        'name': '@',
                        'value': 'a6',
                        'ttl': 10800,
                        'zone_id': '42',
                    },
                ],
                'records': [],
                'error': {
                    'message': 'invalid NS record, invalid NS record, ',
                    'code': 422,
                },
            }),
        ])

        assert result['msg'] == (
            'Errors: Creating NS record "a4" with TTL 10800 for zone 42 failed with unknown reason;'
            ' Creating NS record "a5" with TTL 10800 for zone 42 failed with unknown reason'
        )

    def test_change_change_bad(self, mocker):
        result = self.run_module_failed(mocker, hetzner_dns_record_set, {
            'hetzner_token': 'foo',
            'state': 'present',
            'zone_name': 'example.com',
            'record': 'example.com',
            'type': 'A',
            'ttl': 3600,
            'value': [
                '1.2.3.4.5',
            ],
            '_ansible_remote_tmp': '/tmp/tmp',
            '_ansible_keep_remote_files': True,
        }, [
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/zones', without_query=True)
            .expect_query_values('name', 'example.com')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_LIST_RESULT),
            FetchUrlCall('GET', 200)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records', without_query=True)
            .expect_query_values('zone_id', '42')
            .expect_query_values('page', '1')
            .expect_query_values('per_page', '100')
            .return_header('Content-Type', 'application/json')
            .result_json(HETZNER_JSON_ZONE_RECORDS_GET_RESULT),
            FetchUrlCall('PUT', 422)
            .expect_header('accept', 'application/json')
            .expect_header('auth-api-token', 'foo')
            .expect_url('https://dns.hetzner.com/api/v1/records/125')
            .expect_json_value_absent(['id'])
            .expect_json_value(['type'], 'A')
            .expect_json_value(['ttl'], 3600)
            .expect_json_value(['zone_id'], '42')
            .expect_json_value(['name'], '@')
            .expect_json_value(['value'], '1.2.3.4.5')
            .return_header('Content-Type', 'application/json')
            .result_json({
                'record': {
                    'id': '',
                    'type': '',
                    'name': '',
                    'value': '',
                    'zone_id': '',
                    'created': '',
                    'modified': '',
                },
                'error': {
                    'message': 'invalid A record',
                    'code': 422,
                }
            }),
        ])

        assert result['msg'] == (
            'Error: The updated A record with value "1.2.3.4.5" and TTL 3600 has not been accepted'
            ' by the server with error message "invalid A record" (error code 422)'
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this file contains some public domain test data from
# https://raw.githubusercontent.com/publicsuffix/list/master/tests/test_psl.txt
# The data is marked and documented as public domain appropriately.

# Make coding more python3-ish
from __future__ import absolute_import, division, print_function

__metaclass__ = type


import pytest

from ansible_collections.community.dns.plugins.plugin_utils.public_suffix import (
    PublicSuffixList,
    PUBLIC_SUFFIX_LIST,
)


TEST_GET_SUFFIX = [
    ('', {}, {}, '', ''),
    ('.', {}, {}, '', ''),
    ('foo.com', {}, {}, 'com', 'foo.com'),
    ('bar.foo.com.', {}, {}, 'com.', 'foo.com.'),
    ('BaR.fOo.CoM.', {'normalize_result': True}, {}, 'com.', 'foo.com.'),
    ('BaR.fOo.CoM.', {}, {}, 'CoM.', 'fOo.CoM.'),
    ('com', {}, {}, 'com', ''),
    ('com', {}, {'only_if_registerable': False}, 'com', 'com'),
    ('com', {'keep_unknown_suffix': False}, {}, 'com', ''),
    ('foo.com', {}, {}, 'com', 'foo.com'),
    ('foo.com', {'keep_unknown_suffix': False}, {}, 'com', 'foo.com'),
    ('foobarbaz', {}, {}, 'foobarbaz', ''),
    ('foobarbaz', {}, {'only_if_registerable': False}, 'foobarbaz', 'foobarbaz'),
    ('foobarbaz', {'keep_unknown_suffix': False}, {}, '', ''),
    ('foo.foobarbaz', {}, {}, 'foobarbaz', 'foo.foobarbaz'),
    ('foo.foobarbaz', {'keep_unknown_suffix': False}, {}, '', ''),
    ('-a.com', {}, {}, '', ''),  # invalid domain name (leading dash in label)
    ('a-.com', {}, {}, '', ''),  # invalid domain name (trailing dash in label)
    ('-.com', {}, {}, '', ''),  # invalid domain name (leading and trailing dash in label)
    ('.com', {}, {}, '', ''),  # invalid domain name (empty label)
    ('test.cloudfront.net', {}, {}, 'cloudfront.net', 'test.cloudfront.net'),  # private rule
    ('test.cloudfront.net', {'icann_only': True}, {}, 'net', 'cloudfront.net'),
]


@pytest.mark.parametrize("domain, kwargs, reg_extra_kwargs, suffix, reg_domain", TEST_GET_SUFFIX)
def test_get_suffix(domain, kwargs, reg_extra_kwargs, suffix, reg_domain):
    assert PUBLIC_SUFFIX_LIST.get_suffix(domain, **kwargs) == suffix
    kwargs.update(reg_extra_kwargs)
    assert PUBLIC_SUFFIX_LIST.get_registrable_domain(domain, **kwargs) == reg_domain


# -------------------------------------------------------------------------------------------------
# The following list is taken from https://raw.githubusercontent.com/publicsuffix/list/master/tests/test_psl.txt
# Any copyright for this list is dedicated to the Public Domain. (https://creativecommons.org/publicdomain/zero/1.0/)
# This list has been provided by Rob Stradling of Comodo (see last section on https://publicsuffix.org/list/).
TEST_SUFFIX_OFFICIAL_TESTS = [
    # '' input.
    ('', '', {}),
    # Mixed case.
    ('COM', '', {}),
    ('example.COM', 'example.com', {'normalize_result': True}),
    ('WwW.example.COM', 'example.com', {'normalize_result': True}),
    ('example.COM', 'example.COM', {}),
    ('WwW.example.COM', 'example.COM', {}),
    # Leading dot.
    ('.com', '', {}),
    ('.example', '', {}),
    ('.example.com', '', {}),
    ('.example.example', '', {}),
    # Unlisted TLD.
    ('example', '', {}),
    ('example.example', 'example.example', {}),
    ('b.example.example', 'example.example', {}),
    ('a.b.example.example', 'example.example', {}),
    # Listed, but non-Internet, TLD.
    # ('local', '', {}),
    # ('example.local', '', {}),
    # ('b.example.local', '', {}),
    # ('a.b.example.local', '', {}),
    # TLD with only 1 rule.
    ('biz', '', {}),
    ('domain.biz', 'domain.biz', {}),
    ('b.domain.biz', 'domain.biz', {}),
    ('a.b.domain.biz', 'domain.biz', {}),
    # TLD with some 2-level rules.
    ('com', '', {}),
    ('example.com', 'example.com', {}),
    ('b.example.com', 'example.com', {}),
    ('a.b.example.com', 'example.com', {}),
    ('uk.com', '', {}),
    ('example.uk.com', 'example.uk.com', {}),
    ('b.example.uk.com', 'example.uk.com', {}),
    ('a.b.example.uk.com', 'example.uk.com', {}),
    ('test.ac', 'test.ac', {}),
    # TLD with only 1 (wildcard) rule.
    ('mm', '', {}),
    ('c.mm', '', {}),
    ('b.c.mm', 'b.c.mm', {}),
    ('a.b.c.mm', 'b.c.mm', {}),
    # More complex TLD.
    ('jp', '', {}),
    ('test.jp', 'test.jp', {}),
    ('www.test.jp', 'test.jp', {}),
    ('ac.jp', '', {}),
    ('test.ac.jp', 'test.ac.jp', {}),
    ('www.test.ac.jp', 'test.ac.jp', {}),
    ('kyoto.jp', '', {}),
    ('test.kyoto.jp', 'test.kyoto.jp', {}),
    ('ide.kyoto.jp', '', {}),
    ('b.ide.kyoto.jp', 'b.ide.kyoto.jp', {}),
    ('a.b.ide.kyoto.jp', 'b.ide.kyoto.jp', {}),
    ('c.kobe.jp', '', {}),
    ('b.c.kobe.jp', 'b.c.kobe.jp', {}),
    ('a.b.c.kobe.jp', 'b.c.kobe.jp', {}),
    ('city.kobe.jp', 'city.kobe.jp', {}),
    ('www.city.kobe.jp', 'city.kobe.jp', {}),
    # TLD with a wildcard rule and exceptions.
    ('ck', '', {}),
    ('test.ck', '', {}),
    ('b.test.ck', 'b.test.ck', {}),
    ('a.b.test.ck', 'b.test.ck', {}),
    ('www.ck', 'www.ck', {}),
    ('www.www.ck', 'www.ck', {}),
    # US K12.
    ('us', '', {}),
    ('test.us', 'test.us', {}),
    ('www.test.us', 'test.us', {}),
    ('ak.us', '', {}),
    ('test.ak.us', 'test.ak.us', {}),
    ('www.test.ak.us', 'test.ak.us', {}),
    ('k12.ak.us', '', {}),
    ('test.k12.ak.us', 'test.k12.ak.us', {}),
    ('www.test.k12.ak.us', 'test.k12.ak.us', {}),
    # IDN labels.
    (u'食狮.com.cn', u'食狮.com.cn', {}),
    (u'食狮.公司.cn', u'食狮.公司.cn', {}),
    (u'www.食狮.公司.cn', u'食狮.公司.cn', {}),
    (u'shishi.公司.cn', u'shishi.公司.cn', {}),
    (u'公司.cn', u'', {}),
    (u'食狮.中国', u'食狮.中国', {}),
    (u'www.食狮.中国', u'食狮.中国', {}),
    (u'shishi.中国', u'shishi.中国', {}),
    (u'中国', u'', {}),
    # Same as above, but punycoded.  (TODO: punycode not supported yet!)
    ('xn--85x722f.com.cn', 'xn--85x722f.com.cn', {}),
    ('xn--85x722f.xn--55qx5d.cn', 'xn--85x722f.xn--55qx5d.cn', {}),
    ('www.xn--85x722f.xn--55qx5d.cn', 'xn--85x722f.xn--55qx5d.cn', {}),
    ('shishi.xn--55qx5d.cn', 'shishi.xn--55qx5d.cn', {}),
    ('xn--55qx5d.cn', '', {}),
    ('xn--85x722f.xn--fiqs8s', 'xn--85x722f.xn--fiqs8s', {}),
    ('www.xn--85x722f.xn--fiqs8s', 'xn--85x722f.xn--fiqs8s', {}),
    ('shishi.xn--fiqs8s', 'shishi.xn--fiqs8s', {}),
    ('xn--fiqs8s', '', {}),
]
# End of public domain test data
# -------------------------------------------------------------------------------------------------


@pytest.mark.parametrize("domain, registrable_domain, kwargs", TEST_SUFFIX_OFFICIAL_TESTS)
def test_get_suffix_official(domain, registrable_domain, kwargs):
    reg_domain = PUBLIC_SUFFIX_LIST.get_registrable_domain(domain, **kwargs)
    assert reg_domain == registrable_domain


def test_load_psl_dot(tmpdir):
    fn = tmpdir / 'psl.dat'
    fn.write('''// ===BEGIN BLA BLA DOMAINS===
.com.
// ===END BLA BLA DOMAINS==='''.encode('utf-8'))
    psl = PublicSuffixList.load(str(fn))
    assert len(psl._rules) == 1
    rule = psl._rules[0]
    assert rule.labels == ('com', )
    assert rule.exception_rule is False
    assert rule.part == 'bla bla'


def test_load_psl_no_part(tmpdir):
    fn = tmpdir / 'psl.dat'
    fn.write('''// ===BEGIN BLA BLA DOMAINS===
com
// ===END BLA BLA DOMAINS===
net'''.encode('utf-8'))
    with pytest.raises(Exception) as excinfo:
        PublicSuffixList.load(str(fn))
    assert str(excinfo.value) == 'Internal error: found PSL entry with no part!'



#
# (c) 2019, Ansible by Red Hat, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#

from __future__ import absolute_import, division, print_function

__metaclass__ = type

from textwrap import dedent
from ansible_collections.cisco.ios.tests.unit.compat.mock import patch
from ansible_collections.cisco.ios.plugins.modules import ios_lag_interfaces
from ansible_collections.cisco.ios.tests.unit.modules.utils import (
    set_module_args,
)
from .ios_module import TestIosModule


class TestIosLagInterfacesModule(TestIosModule):
    module = ios_lag_interfaces

    def setUp(self):
        super(TestIosLagInterfacesModule, self).setUp()

        self.mock_get_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network.Config.get_config"
        )
        self.get_config = self.mock_get_config.start()

        self.mock_load_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.network.Config.load_config"
        )
        self.load_config = self.mock_load_config.start()

        self.mock_get_resource_connection_config = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base."
            "get_resource_connection"
        )
        self.get_resource_connection_config = (
            self.mock_get_resource_connection_config.start()
        )

        self.mock_get_resource_connection_facts = patch(
            "ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.resource_module_base."
            "get_resource_connection"
        )
        self.get_resource_connection_facts = (
            self.mock_get_resource_connection_facts.start()
        )

        self.mock_edit_config = patch(
            "ansible_collections.cisco.ios.plugins.module_utils.network.ios.providers.providers.CliProvider.edit_config"
        )
        self.edit_config = self.mock_edit_config.start()

        self.mock_execute_show_command = patch(
            "ansible_collections.cisco.ios.plugins.module_utils.network.ios.facts.lag_interfaces.lag_interfaces."
            "Lag_interfacesFacts.get_lag_interfaces_data"
        )
        self.execute_show_command = self.mock_execute_show_command.start()

    def tearDown(self):
        super(TestIosLagInterfacesModule, self).tearDown()
        self.mock_get_resource_connection_config.stop()
        self.mock_get_resource_connection_facts.stop()
        self.mock_edit_config.stop()
        self.mock_get_config.stop()
        self.mock_load_config.stop()
        self.mock_execute_show_command.stop()

    def test_ios_lag_interfaces_merged(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/1", "mode": "active"}
                        ],
                        "name": "Port-channel11",
                    },
                    {
                        "members": [
                            {"member": "GigabitEthernet0/2", "mode": "active"},
                            {
                                "member": "GigabitEthernet0/3",
                                "mode": "passive",
                            },
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="merged",
            )
        )
        commands = [
            "interface GigabitEthernet0/3",
            "channel-group 22 mode passive",
        ]
        result = self.execute_module(changed=True)
        # print(result["commands"])
        self.assertEqual(result["commands"], commands)

    def test_ios_lag_interfaces_merged_idempotent(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/1", "mode": "active"}
                        ],
                        "name": "Port-channel11",
                    },
                    {
                        "members": [
                            {"member": "GigabitEthernet0/2", "mode": "active"},
                            {"member": "GigabitEthernet0/3", "mode": "active"},
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="merged",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_ios_lag_interfaces_replaced(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/3", "mode": "active"}
                        ],
                        "name": "Port-channel11",
                    },
                    {
                        "members": [
                            {"member": "GigabitEthernet0/1", "mode": "active"},
                            {"member": "GigabitEthernet0/3", "mode": "on"},
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="replaced",
            )
        )
        commands = [
            "interface GigabitEthernet0/1",
            "no channel-group 11 mode active",
            "interface GigabitEthernet0/1",
            "channel-group 22 mode active",
            "interface GigabitEthernet0/3",
            "channel-group 22 mode on",
            "interface GigabitEthernet0/2",
            "no channel-group 22 mode active",
        ]
        result = self.execute_module(changed=True)
        self.assertEqual(result["commands"], commands)

    def test_ios_lag_interfaces_replaced_idempotent(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/1", "mode": "active"}
                        ],
                        "name": "Port-channel11",
                    },
                    {
                        "members": [
                            {"member": "GigabitEthernet0/2", "mode": "active"},
                            {"member": "GigabitEthernet0/3", "mode": "active"},
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="replaced",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_ios_lag_interfaces_overridden(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/2", "mode": "active"},
                            {"member": "GigabitEthernet0/3", "mode": "active"},
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="overridden",
            )
        )

        commands = [
            "interface GigabitEthernet0/1",
            "no channel-group 11 mode active",
            "interface GigabitEthernet0/3",
            "no channel-group 11 mode active",
            "interface GigabitEthernet0/3",
            "channel-group 22 mode active",
        ]
        result = self.execute_module(changed=True)
        print(result["commands"])
        self.assertEqual(result["commands"], commands)

    def test_ios_lag_interfaces_overridden_idempotent(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/1", "mode": "active"}
                        ],
                        "name": "Port-channel11",
                    },
                    {
                        "members": [
                            {"member": "GigabitEthernet0/2", "mode": "active"},
                            {"member": "GigabitEthernet0/3", "mode": "active"},
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="overridden",
            )
        )
        self.execute_module(changed=False, commands=[])

    def test_ios_lag_interfaces_deleted_interface(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(dict(config=[], state="deleted"))
        commands = [
            "interface GigabitEthernet0/1",
            "no channel-group 11 mode active",
            "interface GigabitEthernet0/3",
            "no channel-group 11 mode active",
            "interface GigabitEthernet0/2",
            "no channel-group 22 mode active",
            "interface GigabitEthernet0/4",
            "no channel-group 22 mode active link 20",
            "interface GigabitEthernet0/5",
            "no channel-group 22 link 22",
        ]
        result = self.execute_module(changed=True)
        print(result["commands"])
        self.assertEqual(result["commands"], commands)

    def test_ios_lag_interfaces_deleted(self):
        self.execute_show_command.return_value = dedent(
            """\
            interface Port-channel11
            interface Port-channel22
            interface GigabitEthernet0/1
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/2
             shutdown
             channel-group 22 mode active
            interface GigabitEthernet0/3
             shutdown
             channel-group 11 mode active
            interface GigabitEthernet0/4
             shutdown
             channel-group 22 mode active link 20
            interface GigabitEthernet0/5
             shutdown
             channel-group 22 link 22
            """
        )
        set_module_args(
            dict(config=[dict(name="Port-channel11")], state="deleted")
        )
        commands = [
            "interface GigabitEthernet0/1",
            "no channel-group 11 mode active",
            "interface GigabitEthernet0/3",
            "no channel-group 11 mode active",
        ]
        res = self.execute_module(changed=True)
        self.assertEqual(res["commands"], commands)

    def test_ios_lag_interfaces_parsed(self):
        set_module_args(
            dict(
                running_config=dedent(
                    """\
                    interface Port-channel11
                    interface Port-channel22
                    interface GigabitEthernet0/1
                     shutdown
                     channel-group 11 mode active
                    interface GigabitEthernet0/2
                     shutdown
                     channel-group 22 mode active
                    interface GigabitEthernet0/3
                     shutdown
                     channel-group 11 mode active
                    interface GigabitEthernet0/4
                     shutdown
                     channel-group 22 mode active link 20
                    interface GigabitEthernet0/5
                     shutdown
                     channel-group 22 link 22
                    """
                ),
                state="parsed",
            )
        )
        result = self.execute_module(changed=False)
        parsed_list = [
            {
                "name": "Port-channel11",
                "members": [
                    {"member": "GigabitEthernet0/1", "mode": "active"},
                    {"member": "GigabitEthernet0/3", "mode": "active"},
                ],
            },
            {
                "name": "Port-channel22",
                "members": [
                    {"member": "GigabitEthernet0/2", "mode": "active"},
                    {
                        "member": "GigabitEthernet0/4",
                        "link": 20,
                        "mode": "active",
                    },
                    {"member": "GigabitEthernet0/5", "link": 22},
                ],
            },
        ]
        self.assertEqual(parsed_list, result["parsed"])

    def test_ios_lag_interfaces_rendered(self):
        self.execute_show_command.return_value = dedent(
            """\
            """
        )
        set_module_args(
            dict(
                config=[
                    {
                        "members": [
                            {"member": "GigabitEthernet0/1", "mode": "active"}
                        ],
                        "name": "Port-channel11",
                    },
                    {
                        "members": [
                            {"member": "GigabitEthernet0/2", "mode": "active"},
                            {"member": "GigabitEthernet0/3", "mode": "active"},
                            {
                                "link": 20,
                                "member": "GigabitEthernet0/4",
                                "mode": "active",
                            },
                            {"link": 22, "member": "GigabitEthernet0/5"},
                        ],
                        "name": "Port-channel22",
                    },
                ],
                state="rendered",
            )
        )
        commands = [
            "interface GigabitEthernet0/1",
            "channel-group 11 mode active",
            "interface GigabitEthernet0/2",
            "channel-group 22 mode active",
            "interface GigabitEthernet0/3",
            "channel-group 22 mode active",
            "interface GigabitEthernet0/4",
            "channel-group 22 mode active link 20",
            "interface GigabitEthernet0/5",
            "channel-group 22 link 22",
        ]
        result = self.execute_module(changed=False)
        self.assertEqual(sorted(result["rendered"]), sorted(commands))



#!/usr/bin/python
# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: wafv2_rule_group_info
version_added: 1.5.0
author:
  - "Markus Bergholz (@markuman)"
short_description: wafv2_web_acl_info
description:
  - Get informations about existing wafv2 rule groups.
options:
    state:
      description:
        - Whether the rule is present or absent.
      choices: ["present", "absent"]
      required: true
      type: str
    name:
      description:
        - The name of the rule group.
      required: true
      type: str
    scope:
      description:
        - Scope of wafv2 rule group.
      required: true
      choices: ["CLOUDFRONT","REGIONAL"]
      type: str

extends_documentation_fragment:
- amazon.aws.aws
- amazon.aws.ec2

'''

EXAMPLES = '''
- name: rule group info
  community.aws.wafv2_rule_group_info:
    name: test02
    state: present
    scope: REGIONAL
'''

RETURN = """
arn:
    description: Rule group arn
    sample: arn:aws:wafv2:eu-central-1:11111111:regional/rulegroup/test02/6e90c01a-e4eb-43e5-b6aa-b1604cedf7d7
    type: str
    returned: Always, as long as the web acl exists
description:
    description: Description of the rule group
    sample: Some rule group description
    returned: Always, as long as the web acl exists
    type: str
capacity:
    description: Current capacity of the rule group
    sample: 500
    returned: Always, as long as the rule group exists
    type: int
name:
    description: Rule group name
    sample: test02
    returned: Always, as long as the rule group exists
    type: str
rules:
    description: Current rules of the rule group
    returned: Always, as long as the rule group exists
    type: list
    sample:
        - action:
             allow: {}
          name: eins
          priority: 1
          statement:
              ip_set_reference_statement:
                  arn: arn:aws:wafv2:eu-central-1:111111111:regional/ipset/test02/b6978915-c67b-4d1c-8832-2b1bb452143a
          visibility_config:
              cloud_watch_metrics_enabled: True
              metric_name: fsd
              sampled_requests_enabled: True
visibility_config:
    description: Visibility config of the rule group
    returned: Always, as long as the rule group exists
    type: dict
    sample:
        cloud_watch_metrics_enabled: True
        metric_name: blub
        sampled_requests_enabled: False
"""

try:
    from botocore.exceptions import ClientError, BotoCoreError
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.core import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.ec2 import camel_dict_to_snake_dict
from ansible_collections.community.aws.plugins.module_utils.wafv2 import wafv2_list_rule_groups


def get_rule_group(wafv2, name, scope, id, fail_json_aws):
    try:
        response = wafv2.get_rule_group(
            Name=name,
            Scope=scope,
            Id=id
        )
    except (BotoCoreError, ClientError) as e:
        fail_json_aws(e, msg="Failed to get wafv2 rule group.")
    return response


def main():
    arg_spec = dict(
        state=dict(type='str', required=True, choices=['present', 'absent']),
        name=dict(type='str', required=True),
        scope=dict(type='str', required=True, choices=['CLOUDFRONT', 'REGIONAL'])
    )

    module = AnsibleAWSModule(
        argument_spec=arg_spec,
        supports_check_mode=True
    )

    state = module.params.get("state")
    name = module.params.get("name")
    scope = module.params.get("scope")

    wafv2 = module.client('wafv2')

    # check if rule group exists
    response = wafv2_list_rule_groups(wafv2, scope, module.fail_json_aws)
    id = None
    retval = {}

    for item in response.get('RuleGroups'):
        if item.get('Name') == name:
            id = item.get('Id')

    existing_group = None
    if id:
        existing_group = get_rule_group(wafv2, name, scope, id, module.fail_json_aws)
        retval = camel_dict_to_snake_dict(existing_group.get('RuleGroup'))

    module.exit_json(**retval)


if __name__ == '__main__':
    main()
